package com.maroontress.sanctuary;

import com.maroontress.sanctuary.xclass.XClass;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.util.List;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
   Sanctuaryεư饹Ǥ
*/
public final class Sanctuary {

    /** ХåեΥ */
    private static final int BUFFER_SIZE = 4096;

    /** DTDΥ᥸㡼С */
    private static final int DTD_MAJOR_VERSION = 2;

    /** DTDΥޥʡС */
    private static final int DTD_MINOR_VERSION = 0;

    /** DTDΥե̾ */
    private static final String DTD_FILENAME = "sanctuary.dtd";

    /**
       ư饹Υ󥹥󥹤ޤ
    */
    private Sanctuary() {
    }

    /**
       ѥȥ󥳡ǥ󥰤ꤷƥ饤ޤ

       @param dirName ǥ쥯ȥ̾
       @param fileName ե̾
       @param encoding ϥ󥳡ǥ
       @return 饤
    */
    public static PrintWriter createPrintWriter(final String dirName,
						final String fileName,
						final String encoding) {
	Options opt = Options.getInstance();
	FileOutputStream out;
	Writer wr;
	File dir, file;

	dir = ((dirName != null)
	       ? new File(opt.getOutputDir(), dirName)
	       : opt.getOutputDir());
	dir.mkdirs();
	file = new File(dir, fileName);
	try {
	    out = new FileOutputStream(file);
	} catch (IOException e) {
	    e.printStackTrace();
	    System.exit(1);
	    return null;
	}
	try {
	    wr = new OutputStreamWriter(out, encoding);
	} catch (UnsupportedEncodingException e) {
	    file.delete();
	    System.exit(1);
	    return null;
	}
	return new PrintWriter(new BufferedWriter(wr));
    }

    /**
       jarեΤʤΥեϥǥ쥯ȥ˽Ϥޤ

       @param s ե̾
    */
    public static void copyFile(final String s) {
	InputStream in = Sanctuary.class.getResourceAsStream(s);
	Options opt = Options.getInstance();
	File file = new File(opt.getOutputDir(), s);
	try {
	    FileOutputStream out = new FileOutputStream(file);
	    byte[] data = new byte[BUFFER_SIZE];
	    int size;
	    while ((size = in.read(data)) > 0) {
		out.write(data, 0, size);
	    }
	    out.close();
	    in.close();
	} catch (IOException e) {
	    e.printStackTrace();
	    file.delete();
	    System.exit(1);
	}
    }

    /**
       SanctuaryXMLեɤޤ

       @param pm ѥåޥ͡
       @param xmlFile XMLե
    */
    private static void load(final PackageManager pm, final String xmlFile) {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	factory.setValidating(true);
        try {
	    DocumentBuilder builder = factory.newDocumentBuilder();
	    builder.setEntityResolver(new EntityResolver() {
		public InputSource resolveEntity(final String publicId,
						 final String systemId)
                    throws SAXException {
                    SystemID id = SystemID.parse(systemId);
                    if (id == null) {
                        return null;
                    }
                    if (id.getMajorVersion() != DTD_MAJOR_VERSION
                        || id.getMinorVersion() > DTD_MINOR_VERSION) {
                        throw new SAXException(systemId + ": not supported.");
                    }
		    InputStream in
			= Sanctuary.class.getResourceAsStream(DTD_FILENAME);
		    return new InputSource(in);
		}
	    });
	    builder.setErrorHandler(new ErrorHandler() {
		public void warning(final SAXParseException ex)
		    throws SAXException {
		    fatalError(ex);
		}
		public void error(final SAXParseException ex)
		    throws SAXException {
		    fatalError(ex);
		}
		public void fatalError(final SAXParseException ex)
		    throws SAXException {
		    System.err.println(xmlFile + ":" + ex.getLineNumber()
				       + ":" + ex.getColumnNumber()
				       + ": " + ex.getMessage());
		    System.exit(1);
		}
	    });
	    Document doc = builder.parse(new File(xmlFile));
	    Element root = doc.getDocumentElement();
	    if (root.getNodeName().equals("class")) {
		pm.addGroup(new Class((Element) root, xmlFile));
	    } else if (root.getNodeName().equals("interface")) {
		pm.addGroup(new Interface((Element) root, xmlFile));
	    } else if (root.getNodeName().equals("package")) {
		pm.addPackage(new Package((Element) root));
	    } else if (root.getNodeName().equals("xclass")) {
		pm.addGroup(new XClass((Element) root, xmlFile));
	    }
        } catch (NodeException e) {
	    System.err.println(xmlFile + ": " + e.getMessage());
	    System.err.println(e.getLocation());
	    System.exit(1);
	} catch (Exception e) {
	    System.err.println(xmlFile + ":");
	    e.printStackTrace();
	    System.exit(1);
	}
    }

    /**
       СϤޤ
    */
    private static void version() {
	InputStream in = Sanctuary.class.getResourceAsStream("version");
	byte[] data = new byte[BUFFER_SIZE];
	int size;
	try {
	    while ((size = in.read(data)) > 0) {
		System.out.write(data, 0, size);
	    }
	    in.close();
	} catch (Exception e) {
	    e.printStackTrace();
	}
	System.exit(1);
    }

    /**
       ˡɽޤ
    */
    private static void usage() {
	System.err.print(""
+ "Usage: java com.maroontress.sanctuary.Sanctuary [options] [file...]\n"
+ "Options are:\n"
+ "  --code                  Generate header files.\n"
+ "  --html                  Generate HTML documents.\n"
+ "  --tex                   Generate TeX documents.\n"
+ "  --stub                  Generate stubs.\n"
+ "  --dtd                   Generate DTD file (" + DTD_FILENAME + ").\n"
+ "  --without-c++           Generate header files not for C++.\n"
+ "  --with-base-xclass=FQCN Specify the default base xclass.\n"
+ "  --input-file=FILE       Read the list of files from FILE.\n"
+ "  --output-dir=DIR        Specify where to place generated files.\n"
+ "  --version               Show version.\n"
+ "");
	System.exit(1);
    }

    /**
       Sanctuary¹Ԥޤ

       @param av ޥɥ饤󥪥ץ
    */
    public static void main(final String[] av) {
	boolean printCode = false;
	boolean printHtml = false;
	boolean printTex = false;
	boolean printStub = false;
	boolean printDtd = false;
	PackageManager pm = PackageManager.getInstance();
	Options opt = Options.getInstance();
	ArgumentParser ap = new ArgumentParser(av);
	Argument arg;

	while ((arg = ap.getArgument()) != null && arg.isOption()) {
	    String name = arg.getName();
	    String value = arg.getValue();
	    if (name.equals("--version")) {
		version();
	    } else if (name.equals("--html")) {
		printHtml = true;
	    } else if (name.equals("--code")) {
		printCode = true;
	    } else if (name.equals("--tex")) {
		printTex = true;
	    } else if (name.equals("--stub")) {
		printStub = true;
	    } else if (name.equals("--dtd")) {
		printDtd = true;
	    } else if (name.equals("--without-c++")) {
		opt.setCXXHeaderSupport(false);
	    } else if (name.equals("--with-base-xclass") && value != null) {
		opt.setBaseXClass(value);
	    } else if (name.equals("--output-dir") && value != null) {
		opt.setOutputDir(value);
	    } else if (name.equals("--input-file") && value != null) {
		opt.setInputFile(value);
	    } else {
		usage();
	    }
	}
	while (arg != null) {
	    String name = arg.getName();
	    load(pm, name);
	    arg = ap.getArgument();
	}
	File inputFile = opt.getInputFile();
	if (inputFile != null) {
	    BufferedReader rd;
	    String name;
	    try {
		rd = new BufferedReader(new FileReader(inputFile));
		while ((name = rd.readLine()) != null) {
		    load(pm, name);
		}
	    } catch (FileNotFoundException e) {
		System.err.println("File not found: " + e.getMessage());
		System.exit(1);
	    } catch (IOException e) {
		e.printStackTrace();
		System.exit(1);
	    }
	}
	if (printDtd) {
	    copyFile(DTD_FILENAME);
	}
	if (printHtml) {
	    pm.printIndexHtml();
	    copyFile("sanctuary.css");
	}
	if (printTex) {
	    pm.printSanctuaryTex();
	    copyFile("sanctuary.sty");
	}
	List<Group> v = pm.groupList();
	for (Group g : v) {
	    if (printCode) {
		g.printCode();
	    }
	    if (printHtml) {
		g.printHtml();
	    }
	    if (printTex) {
		g.printTex();
	    }
	    if (printStub) {
		g.printStub();
	    }
	}
    }
}
